/* File yasep/core-js/yasep_asm.js
This file is distributed under terms of the Affero GPL v3 license or later, see http://yasep.org
created 2005-06-12 by whygee@f-cpu.org 
2006-08-10 : populated the file (at last)
2006-08-10 : finished the parser and added encoding table/functions (without the flags though)
2006-09-02 : moved to vspsim/JScore/
2006-04-01 : added emit_asm(), set unused register fields,
  tried to enhance the pseudo-instruction parsing
2007-04-04 : update for the Dx pseudo-instructions
2007-04-06 : added check_imm16(i) and 5 new forms for jump/skip
2007-04-08 : cleanup with new JMP/SKIP instruction formats
2007-04-09 : changed the argument for the encoders (form -> flag)
    added the '?' operand, and the X/RX/IX/QX/RRX/IRX/QRX encoders
2007-04-17 : encoded_ignore_fields turned into encoded_fields etc.
2007-04-18 : restricted the range of Imm16 for some instructions
2007-11-05 : rebranding to YASEP
2007-12-25 : MOV 0 -> CLR optimisation 
2008-11-17 : added OPTIONAL|YASEP32_ONLY|YASEP16_ONLY
2009-07-15 : updates
2009-07-19 : small fixes
2009-07-27 : small mask bug with FORM_R,
 objectified the assembler to reduce namespace clutter in the global object.
2009-07-28 : added Y
2009-07-28 : ALIAS_RIR, IgnoreImmSign
2009-07-30 : si4/snd
2009-08-10 : swapped snd & si4 for FORM_RIR, removed SWAPDEST
2009-08-15 : replaced RIR with IRR, added RRR and iRR
2009-08-16 : adding conditional codes
2009-08-23 : removed FORM_X
2011-09-08 : back to work : changed the object structure, create an assembly context,
                put the add_message() as an external function pointer
2011-09-17 : solving the NOP bug by defining NOP=MOV NPC NPC and long NOP forbidden.
               Trying to add IMM20 to MOV
2011-10-17 : bug with CALL and CALL2, IMM size troubles
2011-11-30 : moved stuff to yasep_forms.js, removed the ASM object
2012-09-21 : adding DU
2012-11-23 : report the condition name to the context (lets BIT0/BIT1 and others be configured)

This file does the assembly (encoding) of instructions
both for interactive debug and coding. It takes
a string of text and spits out an encoded integer
(and other informations) in an object.

 dependency:
   yasep/JScore/txt_int.js,
//////////////  (yasep/JSgui/yasep_messages.js+handling [define an empty function if undesired]),
   yasep/JScore/yasep_forms.js/fields/flags/aliases/conditions
   yasep/JScore/yasep_opcodes.js+children (group_eu.js etc.)
*/

// for parsing DU:
ASCIIescapes={
  '"':34,
  '\\':92,
  '/':47,
  'b':8,
  'f':12,
  'n':10,
  'r':13,
  't':9
};

// This creates the object that is passed to and from the assembler
Y.ASM_context=function(add_message, emit_bin, silent, core) {
  // input
  if (typeof add_message=="function")
    this.add_message = add_message;
  else
    this.add_message = emptyFunc;// function() {};
  this.YASEP_TYPE=32; // YASEP16/YASEP32 ?

  if (emit_bin)
    this.emit_bin=emit_bin;
  else
    this.emit_bin=emptyFunc;

  if (typeof silent == "boolean")
    this.silent=silent;
  else
    this.silent=false;

  if (typeof core == "object") // it must be profiles["string"] directly
    this.core=core;
  else
    this.core=false; // ugly

  // output :
/* later in the encoder function:
  // preliminary cleanup
  t.encoded_fields=0;       // mask of the used/useful bits
  t.encoded_instruction=0;  // The result of encoding
  t.encoded_size_bytes=0;   // 2 if short form, 4 if long form
  t.asm_operands=[];        // used to pass parameters to external functions
  t.pseudo_flag=0;          // set to 1 if DB/DH/DW is being decoded
  t.X=0;                    // instruction is extended
  t.form=""
*/
};


//////////////////////////////////////////////////////////
/////////////////////// The parser ///////////////////////

// Syntax of an instruction line :
// opcode (operands)* (condition (operands*)) (; comment)
// exception : DU "st\"ring"

Y.encode_instruction=function(input_text,t) {
   // arguments : string to encode, ASM context (asm=new ASM_context(add_message))
   // return value : true=OK, false=error

  // preliminary cleanup
  t.asm_operands=[];
  t.encoded_fields
    =t.encoded_instruction
    =t.encoded_size_bytes
    =t.pseudo_flag
    =t.X
    =0;
  t.name
   =t.form
   =t.to
   =t.cond
   ="";

  var a, b, c, i, f=0, g=0, k, I=false, //I=true when Imm16
      n, p="", r, to, Ytype,//YASEP type reset to ",

  // lazy preprocessing
  // Character strings are not processed yet !
  s=input_text.toUpperCase();   // the original is preserved for later if it is DU
  s=s.replace(/[;].*/,"");  // remove any trailing comment
  s=s.replace(/[,]/g," ");  // remove the commas
  s=s.replace(/\s+/g," ");  // combine multiple spaces
  s=s.replace(/(^[ ])|([ ]$)/g,""); // strip leading or trailing garbage

  // split the string into "tokens"
  a=s.split(" ");
  n=a.length;

  if (a[0]=="") {
    if (t.silent)
      return true;
    t.add_message("red", I8N("R")+I8N("asmie"));
    return false;
  }

  /* opcode lookup */
  if (Y.table_opcode[a[0]] == undefined) {

    // Opcode not found. Try a pseudo-instruction :
    t.encoded_size_bytes=0;
    var bytes=Y.table_pseudo_instructions[a[0]];
    if (typeof bytes!="number") {
      t.add_message("red", I8N("R")+I8N("asmuo")+a[0]);
      return false
    }
    else {
      Ytype=I8N("asmty");

      switch (a[0]) {
        case "YASEP":
          t.YASEP_TYPE=0;
          t.add_message("green", Ytype+"default.");
          return true;

        case "YASEP16":
          t.YASEP_TYPE=16;      /////////////////////// TODO : accept profile names ///////////////////////
          t.add_message("green", Ytype+"16-bit.");
          return true;

        case "YASEP32":
          t.YASEP_TYPE=32;
          t.add_message("green", Ytype+"32-bit.");
          return true;
      } //  default: process DB/DH/DW/DU pseudo-instructions, that's where we use the "bytes" variable:
      t.pseudo_flag=1;
      t.encoded_instruction=0;

      if (n < 2) {
        t.add_message("red", I8N("R")+I8N("asmmd"));
        return false
      }

//DU
      if (a[0]=="DU") { // parse a UTF8 string, thanks JSON.org :-) / http://www.json.org/string.gif

        // skip past the leading quotation mark
        if((i=input_text.indexOf('"'))<0) {
          t.add_message("red", I8N("R")+I8N("asmqm"));
          return false
        }
        i++;
        while (i<input_text.length){
          c=input_text[i++];
          if (c=='"'){
// TODO : scan for garbage after this
            t.add_message("blue"," OK, "+t.encoded_size_bytes+I8N("asmlb"));
            return true;
          }
          if (c=='\\'){
            // lookup the next char
            if(typeof(f=input_text[i])=="undefined")
              break; // error
            // is it an escape code ?
            if(typeof(f=ASCIIescapes[f])=="number"){
              t.emit_bin(f, 1);
              t.encoded_size_bytes++;
              i++;
              continue;
            }
            else{
              t.add_message("red", I8N("R")+"\\"+i+" "+I8N("asmiec"));
              return false;
            }
          }
          // turn the character into UTF8 bytes
          c=unescape(encodeURIComponent(c));
          for (b in c){
            t.emit_bin(c.charCodeAt(b), 1);
            t.encoded_size_bytes++;
          }
        }
        t.add_message("red", I8N("R")+I8N("asmqm"));
        return false;
      }

// DB, DH, DW:
      // scan the operand list and tries to convert the numbers
      for (i=1; i<n; i++) {
        r=str2int(a[i]); // quiet : don't send an error message
        if (r==null) {
          t.add_message("red", I8N("R")+I8N("asmin")+a[i]);
          return false
        }

        // strip the MSB
        b=0;
        c=bytes;
        while (c>0) { // stupid way to create a stupid mask
          b= (b<<8) | 0xFF;
          c--;
        }
        b&=r;
        if (b!=r)
          t.add_message("orange", I8N("W")+I8N("asmln")+a[i]+I8N("asmitl")+(bytes*8)+I8N("asmbk"));

//        if (typeof t.emit_bin == "function")
          t.emit_bin(b, bytes);

        // This works for data <= 32 bits total :
        if (t.encoded_size_bytes<4)
          t.encoded_instruction+= (b<<(t.encoded_size_bytes*8));
        t.encoded_size_bytes+=bytes;
      }
      t.add_message("blue"," OK, "+t.encoded_size_bytes+I8N("asmlb"));

    } // else
  }
  // normal instruction :
  else {
    if (n > 7) {
      // Should be enough for a complex extended instruction like ADD R1 R2 R3 LSB0 R4
      t.add_message("red", I8N("R")+I8N("asmtma"));
      return false
    }

//    t.name=a[0];
//    to=Y.table_opcode[a[0]];
    to=Y.table_opcode[t.name
                      =a[0]];
    t.encoded_instruction=to.val;
    t.encoded_fields=0xFF;    // Y.FIELD_OPCODE|Y.FIELD_EXTFORM|Y.FIELD_IMMREG;
    t.encoded_size_bytes=2; // Default. size=4 is set in check_imm16() and force_extended_form()
    f=to.forms;
    g=to.flags;
    if (g & Y.READ_DST)
      Y.force_extended_form(t);

    if (t.core && t.core.valid_instructions[Y.opcode_table[to.val].name] != 1) {
      t.add_message("orange", I8N("R")+I8N("asmcpo"));
      return false
    }

// Parse the operands and build the form :

    // decode the other fields (speculatively)
    for (i=1; i<n; i++) {
      // first try with a register
      r=Y.names_registers[a[i]];
      if (typeof r == "number") {
        p+="R";
        t.asm_operands[i]=r;
      }
      else {
        // is it a number ?
        r=str2int(a[i]); // quiet : don't send an error message
        if (typeof r == "number") {
/*
          c="I";
          if (((r >= -8) && (r <= 7))
          ||  ((r >=  0) && (r <= 15) && (g&Y.IgnoreImmSign))
          && (f & Y.FORMS_i))
            c="i";
          else
            I=true;
          p+=c;
*/
          p+="I"; // will be changed later
          t.asm_operands[i]=r;
        }
        else {
          // process the conditions
          if (k=Y.tableConditionAll[a[i]]) {
/*            if (I==true) {
              t.add_message("red",I8N("param")+i+", "+I8N("R")+I8N("asmchl"));
              return false
            }
*/
            if (g & Y.NO_CONDITION)
              t.add_message("orange",I8N("W")+I8N("asmiic"));

            Y.force_extended_form(t);
            t.encoded_instruction|= k[0]<<Y.OFFSET_CNDCODE;
            r=k[1];
            t.cond=a[i];

            if (r>0) {
              i++;
              if (r==1) {
                r=Y.names_registers[a[i]]; // Expect a register
                if (typeof r == "number")
                  Y.encode_src4(r,t);
                else {
                  t.add_message("red", I8N("R")+I8N("asmrop"));
                  return false
                }
              }
              else { // (r==2)
                r=str2int(a[i]);
                if (typeof r == "number") { // Expect a number
                  if ((r<0) || (r>16)) {
                    t.add_message("red", I8N("R")+I8N("asmc15"));
                    return false
                  }
                  Y.encode_src4(r,t);
                }
                else {
                  t.add_message("red", I8N("R")+I8N("asmnie"));
                  return false
                }
              }
            }
            // complain if there is stuff after this :
            if (i+1 != n) {
              t.add_message("red", I8N("R")+I8N("asmtmo"));
              return false
            }
          } // condition
          else {
            t.add_message("red", I8N("R")+I8N("asmio")+a[i]);
            return false
          }
        }
      }
    } // end for() loop

    if (p=="") // no operand ?
      p="FORM_ALONE";
    else
      p="FORM_"+p;

    if (!Y.TableEncoders[p]) {
      // try with a long immediate
      p=p.replace(/i/g,"I"); // Note : it may not work if more than 1 immediate field is implemented in YASEP
      if (!Y.TableEncoders[p]) {
        t.add_message("red", I8N("R")+I8N("asmoc")+p);
        return false
      }
    }

    if ((t.YASEP_TYPE==32) && (g & Y.YASEP16_ONLY)) {
      t.add_message("orange", I8N("W")+a[0]+I8N("asmo16"));
    } else {
      if ((t.YASEP_TYPE==16) && (g & Y.YASEP32_ONLY)) {
        t.add_message("orange", I8N("W")+a[0]+I8N("asmo32"));
      }
    }

    // Now we have one entry in the bitmask, check if it is present in the opcode
    if (!(f & Y[p])) { // form lookup
      t.add_message("orange", I8N("R")+p+I8N("asmopc")+a[0]);
      return false
    }

    // Check that the operands are OK
    if ((typeof to.check=="function")
           && (!to.check(t,a,p)))
         return false;

    // call the form-specific encoder
    if (!Y.TableEncoders[p].en(g,t)) // relegate the rest of the work to specific routines in yasep_forms.js
      return false;

    // Verify that something has been really processed
    if (~(t.encoded_fields) & t.encoded_instruction)
       t.add_message("orange",I8N("W")+I8N("asmib")
        + " encoded_fields="+int2hex(t.encoded_fields)
        +" instruction="+int2hex(t.encoded_instruction));

    t.form=p;
    t.ot=Y.opcode_table[to.val]; // the instruction is OK (also predecodes for the simulator)

    // the end :
    t.add_message("blue"," OK: opcode="+Y.decorate_opcode(t.encoded_instruction & Y.FIELD_OPCODE)
         +" form="+p+" "+t.encoded_size_bytes+" bytes mask="
         +int2hex(t.encoded_fields,(t.encoded_size_bytes*2)-1,(t.encoded_size_bytes*2)-1));
//    if (typeof t.emit_bin == "function")
      t.emit_bin(t.encoded_instruction, t.encoded_size_bytes);
  } // instruction
  return true
}
