/* yasep/core-js/yasep_forms.js
This file is distributed under terms of the Affero GPL v3 license or later, see http://yasep.org
created 2006-08-10 by whygee@f-cpu.org
2006-08-12
2006-09-02 : moved to vspsim/JScore/
2007-04-01 : added FORM_RI
2007-04-05 : added the jmp/skip related forms
2007-04-08 : cleanup
2007-04-09 : moved elements from JSgui/decoration.js
   to keep symbols in sync, moving the flags to vsp_flags.js
2007-11-05 : rebranding to YASEP, added support for flags IMM16_5LSB and IMM16_8LSB
2009-07-15 : cleanup for the current architecure
2009-07-27 : new version with Y
2009-07-30 : linked to instructions.html
2009-08-16 : extended forms
2009-08-23 : removed FORM_X, off-by-one in form2html
2011-09-17 : added IMM20
2011-11-30 : moving stuff from core-js/yasep_asm.js
2012-03-18 : changed the IMM20 condition so only used 
             in YASEP32, to have more cleared bits in the mask
2012-09-25 : added IMM_LSB
2013-07-21 : FORM_I added (though unused again)
2013-08-10 : PC_Imm4 and Imm4_PC added
2014-02-06 : discarded the funky PC thing, renamed/moved some variables/fields, dropped FORM_RI & FORM_Ri etc.
2014-02-13 : FORM_I is gone too now.

The "form" of an instruction describes the list and type of fields that follow the opcode.
 note : 'forms' are represented as bitfields, so their individual values are powers of 2.

Dependency : none
*/

var Y={ // declare the Y object

  TableForm:[],
  MAX_FORM:1,

  random_register:function(r1,r2){
    var c;
    do
      c=Y.registers_names[intrnd(16)];
    while((c===r1)||(c===r2)); // lazy and slow algorithm
    return c;    
  },

  random_Imm4:function(f){
    var m=7;
    if (f & Y.IMM_LSB)
      m=3; // assume YASEP32
    return intrnd(m)+1;
  },

  random_Imm16:function(f){
    var a=256, b=32768-a;
    if (f & Y.IMM20)  // assume YASEP32
      b=524288-a;
    else {
      if (f & Y.IMM16_5LSB)
        a=b=16;
      else {
        if (f & Y.IMM16_8LSB) {
          a=32;
          b=256-a;
        }
      }
    }
    return a+intrnd(b);
  },

  /////////////////////// Routines for the assembler ///////////////////////

  encode_snd:function(data,t){
    t.encoded_instruction |= data<<Y.OFFSET_SND;
    t.encoded_fields|=Y.FIELD_SND;
    return true
  },

  encode_si4:function(data,t){
    t.encoded_instruction |= data<<Y.OFFSET_SI4;
    t.encoded_fields|=Y.FIELD_SI4;
    return true
  },

  encode_dst:function(data,t){
    t.encoded_instruction |= data<<Y.OFFSET_DST;
    t.encoded_fields|=Y.FIELD_DST;
    return true
  },

  encode_src4:function(data,t){
    t.encoded_instruction |= data<<Y.OFFSET_CND;
    return true
  },

  check_imm16:function(i,flags,t) {
    var a=t.asm_operands[i],
        c=16;
    // Some instructions don't use the whole field
    if (flags & Y.IMM16_5LSB) {
      if (t.YASEP_TYPE==16) // SHL
        c=4;
      else
        c=5;
    }
    else {
      if (flags & Y.IMM16_8LSB)
        c=8;
/*      else {
        if (flags & Y.IMM_LSB) {
          if (t.YASEP_TYPE==16) // IE
            c=1;
          else
            c=2;
        }
      }*/
    }
    // Add the 4 MSB if IMM20
    if ((flags & Y.IMM20) && (t.YASEP_TYPE!=16)) {
      Y.encode_snd((a >> 16)&15, t);
      c = 20;
    }

    var d=(1<<c)-1, // build the bitmask
        b=a&d; // trim the value
    t.encoded_instruction |= 1 | (b<<16);
    t.encoded_fields|=d<<12; // OFFSET_SI4
    t.encoded_size_bytes=4;

    if (((a>0)&&(a>d))  // check that no MSB was discarded
    ||  ((a<0)&&(a<-((d+1)>>1))))
      t.add_message("orange", I8N("W")+"Imm16 "+I8N("lsbx")+c+I8N("lsbx2"));
    return true
  },

  check_i:function(a,flags,t) {

// Imm4_PC: the other register fields (and destination) must be already filled
    if ((t.X==1) && !(t.encoded_instruction & Y.FIELD_SND)) { // DST==0
      t.add_message("green", "Imm4: "+a +"=>"+Y.Imm4_PC[a]);
      a = Y.Imm4_PC[a];
      if (typeof a != "number") {
        t.add_message("red", "Wrong number (can't translate operand for PC");
        return false
      }
    }

    var b,c=4,d;
    if (flags & Y.IMM_LSB) {
      if (t.YASEP_TYPE==16) // IE
        c=1;
      else
        c=2;
    }
    d=(1<<c)-1;
    b=a&d;

    if (((a>0)&&(a>d))  // check that no MSB was discarded
    ||  ((a<0)&&(a<-((d+1)>>1))))
      t.add_message("orange", I8N("W")+"Imm4 "+I8N("lsbx")+c+I8N("lsbx2"));

    t.encoded_fields      |= d <<Y.OFFSET_SI4;// Y.FIELD_SI4;
    t.encoded_instruction |= b <<Y.OFFSET_SI4
      | ((t.X==1) ? Y.FIELD_AUX : Y.FIELD_EXTFORM);
    return true;
  },

////////////////////////
  force_extended_form:function(t){
    t.X=1; // set the extended flag
    t.encoded_instruction |= 3;
    t.encoded_size_bytes=4;
    t.encoded_fields |= Y.FIELD_AUX | Y.FIELD_CNDCODE | Y.FIELD_CND;
  },


  TableEncoders:{

////////////////// Short register forms are declared first //////////////////

// No immediate : NOP, HALT, INV
    "FORM_ALONE":{
      ex:function() {
        return ""
      },
      en:function() {
        /* nothing */
        return true
      },
      de: function() {
        return "FORM_ALONE_F, AO_0()";
      },
      lsb: function() {
        return 0;
      }
    }, // ok

// alias for some FORM_RR : NEG NOT
    "FORM_R":{
      ex:function() {
        return Y.random_register()
      },
      en:function(f,t) {
        Y.encode_snd(t.asm_operands[1],t);
        if (f & Y.ALIAS_RR) {
          // used by :
          // NEG = 0 - snd
          // NOT = 0 | ~snd
          if (t.X==1)
            Y.encode_dst(t.asm_operands[1],t);
          else
            Y.encoded_instruction |= 2; // force FORM_iR
          Y.check_i(0,f,t);
        }
        else
          // IPE : just writes to a register
          if (f & Y.NO_WRITEBACK) {
            t.add_message("red", "Sorry, don't know how to encode this instruction");
            return false
          }
        return true
      },
      de: function() {
        return "FORM_iR_F   , AO_1(OP_REG) /*!*/";
      },
      lsb: function() {
        return 2; // iR
      }
    }, // ok

// used by ROP2, ASU, SHL... 
// ADD AND ANDN CALL CMPS CMPU ESB ESH EZB EZH GET IB IH IHH IN IPC IPE IPR
// LUT8 MOV MUL8H MUL8L MULI NAND NEG NOR NOT OR ORN OUT PUT ROL ROR
// SAR SHL SHR SMAX SMIN SUB UMAX UMIN XOR XORN
    "FORM_RR":{
      ex:function(x,y){
        var a,b="";
        if (y=="CALL")
          b="PC";
        a=Y.random_register(b);
        return a+" "+Y.random_register(a,b)
      },
      en:function(f,t) {
        // NEG = 0 - src1
        // NOT = 0 | ~src1
        if (f & Y.ALIAS_RR) {
          Y.force_extended_form(t);
          Y.encode_snd(t.asm_operands[1],t);
          Y.encode_dst(t.asm_operands[2],t);
          Y.check_i(0,f,t);
        }
        else {
          Y.encode_si4(t.asm_operands[1],t);
          if (t.X==1)
            Y.encode_dst(t.asm_operands[2],t);
            Y.encode_snd (t.asm_operands[2],t);
        }
        return true
      },
      de: function(f) {
        if (f & Y.ALIAS_RR)  // NOT, NEG
          return "FORM_iRR_F  , AO_2(OP_REG,OP_REG) /*!*/";
          return "FORM_RR_F   , AO_2(OP_REG,OP_REG)";
      },
      lsb: function(f) {
        if (f & Y.ALIAS_RR)  // NOT, NEG
          return 3; // iRR
        return 0; // RR
      }
    }, // ok

////////////////// Short immediate forms //////////////////

    // CRIT
    "FORM_i":{
      ex:function(){
        return Y.random_Imm4()
      },
      // used only by CRIT which is not abortable and has no destination
      // so no need to deal with the extended form.
      en:function (f,t) {
        return Y.check_i(t.asm_operands[1],f,t)
      },
      de: function() {
        return "FORM_i_F    , AO_1(OP_IMM4)";
      },
      lsb: function() {
        return 2;
      }
    }, // ok

    // ADD AND ANDN CALL CMPS CMPU GET IB IH IHH LUT8 MOV MUL8H MUL8L
    // MULI NAND NOR OR ORN ROL ROR SAR SHL SHR SMAX SMIN SUB UMAX UMIN XOR XORN
    "FORM_iR":{
      ex:function(x,y){
        var b="";
        if (y=="CALL")
          b="PC";

        return Y.random_Imm4()+" "+Y.random_register(b)
      },

      en:function (f,t,X) {
        if (t.X==1)
          Y.encode_dst(t.asm_operands[2],t);
        Y.encode_snd(t.asm_operands[2],t);
        return Y.check_i(t.asm_operands[1],f,t)
      },
      de: function() {
        return "FORM_iR_F   , AO_2(OP_IMM4,OP_REG)";
      },
      lsb: function() {
        return 2;
      }
    }, // ok
/*
    // IN IPC IPR OUT PUT
    "FORM_Ri":{
      ex:function(){
        return Y.random_register()+" "+Y.random_Imm4()
      },

      en:function (f,t) {
        // no need to deal with dst
        // FIXME : Ri non extendable ? abortable ?
        return (Y.encode_snd(t.asm_operands[1],t)
                && Y.check_i(t.asm_operands[2],f,t))
      },
      de: function() {
        return "FORM_Ri_F   , AO_2(OP_REG,OP_IMM4)";
      },
      lsb: function() {
        return 2;
      }
    }, // ok
*/
////////////////// Long immediate forms //////////////////

    // Long immediate :
    // ADD AND ANDN CALL2 LUT8 MUL8H MUL8L NAND NOR OR ORN
    // ROL ROR SAR SHL SHR SMAX SMIN SUB UMAX UMIN XOR XORN
    "FORM_IRR":{
      ex:function(f){
        var a=Y.random_register();
        return Y.random_Imm16(f)+" "+a+" "+Y.random_register(a)
      },
      en:function (f,t) {
        if(!Y.check_imm16(1,f,t))
          return false;
               Y.encode_snd(t.asm_operands[2],t);
        return Y.encode_si4(t.asm_operands[3],t)
      },
      de: function(f) {
        var i="FORM_IRR_F  , AO_3(OP_IMM";
        if (f & Y.IMM16_5LSB)
          return i+"5 ,OP_REG,OP_REG)";
        if (f & Y.IMM16_8LSB)
          return i+"8 ,OP_REG,OP_REG)";
        return   i+"16,OP_REG,OP_REG)";
      },
      lsb: function() {
        return 1;
      }
    }, // ok

    // SB, and alias for some FORM_IRR
    // ADD AND ANDN CALL CMPS CMPU GET IB IH IHH LUT8 MOV MUL8H MUL8L MULI
    // NAND NOR OR ORN ROL ROR SAR SHL SHR SMAX SMIN SUB UMAX UMIN XOR XORN
    "FORM_IR":{
      ex:function(f,y){
        var b="";
        if (y=="CALL")
          b="PC";
        return Y.random_Imm16(f)+" "+Y.random_register(b)
      },
      en:function (f,t) {
        if(!Y.check_imm16(1,f,t))
          return false;
        var u=t.asm_operands[2];
        if (f & Y.SWAP_IR) // GET, MOV
          return Y.encode_si4(u,t);
        if (f & Y.ALIAS_IRR) // ADD...
               Y.encode_si4(u,t);
        return Y.encode_snd(u,t)
      },
      de: function(f) {
        var i="FORM_IR_F   , AO_2(OP_IMM";
        if (f & Y.IMM20)
          return i+"20,OP_REG)";
        if (f & Y.IMM16_5LSB)
          return i+"5 ,OP_REG)";
        if (f & Y.IMM16_8LSB)
          return i+"8 ,OP_REG)";
        return   i+"16,OP_REG)";
      },
      lsb: function() {
        return 1;
      }
    }, // ok

/*    // HALT
    "FORM_I":{
      ex:function(f,y){
        return Y.random_Imm16(f)
      },
      en:function (f,t) {
        return Y.check_imm16(1,f,t);
      },
      de: function(f) {
        return "FORM_I_F   , AO_1(OP_IMM16)"
      },
      lsb: function() {
        return 1
      }
    }, // ok
    // IN IPC IPE IPR OUT PUT
    "FORM_RI":{
      ex:function(f){
        return Y.random_register()+" "+Y.random_Imm16(f)
      },
      en:function (f,t) {
        return (Y.check_imm16(2,f,t)
             && Y.encode_snd(t.asm_operands[1],t))
      },
      de: function(f) {
        var i="FORM_RI_F   , AO_2(OP_REG,";
        if (f & Y.IMM20)
          return i+" OP_IMM20)";
        if (f & Y.IMM16_5LSB)
          return i+" OP_IMM5)";
        if (f & Y.IMM16_8LSB)
          return i+" OP_IMM8)";
        return   i+"OP_IMM16)";
      },
      lsb: function() {
        return 1;
      }
    }, // ok
*/
////////////////// Extended forms //////////////////

    // ADD AND ANDN CALL2 IB IH IHH LUT8 MUL8H MUL8L
    // NAND NOR OR ORN ROL ROR SAR SHL
    // SHR SMAX SMIN SUB UMAX UMIN XOR XORN
    "FORM_RRR":{
      ex:function(){
        var a=Y.random_register();
        var b=Y.random_register(a);
        return a+" "+b+" "+Y.random_register(a,b);
      },

      en:function(f,t) {
        Y.force_extended_form(t);
        Y.encode_si4 (t.asm_operands[1],t);
        Y.encode_snd (t.asm_operands[2],t);
        Y.encode_dst (t.asm_operands[3],t);
        return true
      },
      de: function() {
        return "FORM_RRR_F  , AO_3(OP_REG,OP_REG,OP_REG)";
      },
      lsb: function() {
        return 3;
      }
    }, // ok

    // ADD AND ANDN CALL2 IB IH IHH LUT8 MUL8H MUL8L NAND NOR OR
    // ORN ROL ROR SAR SHL SHR SMAX SMIN SUB UMAX UMIN XOR XORN
    "FORM_iRR":{
      ex:function(f){
        var a=Y.random_register();
        return ( (a==0)? Y.PC_Imm4[intrnd(16)] : Y.random_Imm4(f))
            + " "+a+" "+Y.random_register(a)
      },
      en:function(f,t) {
        Y.force_extended_form(t);
        Y.encode_snd (t.asm_operands[2],t);
        Y.encode_dst (t.asm_operands[3],t);
        return Y.check_i(t.asm_operands[1],f,t)
      },
      de: function(f) {
        return "FORM_iRR_F  , AO_3(OP_IMM"
          +(f & Y.IMM_LSB)?2:4
          +",OP_REG,OP_REG)";
      },
      lsb: function() {
        return 3;
      }
    } // ok

  }, // TableEncoders


  form2html:function(form,t,n){
    var j=1, k, a="";
    if (!n)
      n=0; // set n to -32 to avoid carriage returns, forget it in the opcode map.

    do {
      if (form & j) {
        k=Y.TableForm[j];
        k=k.slice(5); //k.replace(/FORM_/,"");

        n++; // manage the small Instruction Table spaces
        if((n>3)&&(t==1)) {
          a+=",<br>";
          n=1;
        }
        else
          if(n>1)
            a+=",";

        a+='<a class="innerLink" onclick="wa(\'doc/forms#FORM_'+k+'\')" href="#!doc/forms#FORM_'+k+'">'+k+"</a>";
      }
      j<<=1;
    } while (j<=form);

    return a;
  }

}; // Y


for (var i in Y.TableEncoders) {
  blackList(i);
  Y.TableForm[Y[i]=Y.MAX_FORM]=i;
  Y.MAX_FORM*=2;
};

Y.FORMS_R = Y.FORM_ALONE | Y.FORM_R | Y.FORM_RR;
Y.FORMS_i = Y.FORM_i|Y.FORM_iR;
Y.FORMS_I = Y.FORM_IRR|Y.FORM_IR;
Y.FORMS_EXTENDED = Y.FORM_RRR | Y.FORM_iRR;
Y.FORMS_EXTENDABLE = Y.FORMS_EXTENDED | Y.FORMS_i | Y.FORMS_R;
// not extendable : Y.FORMS_I*


/////////////////////////////// Imm4 LUT for operation with PC : ///////////////////////////////

Y.Imm4_PC={};
Y.PC_Imm4={};
/* ={
  0: 16,
  1: 18,
  2: 20,
  3: 22,
  4:  8,
  5: 10,
  6: 12,
  7: 14,
  8:-16,
  9:-14,
 10:-12,
 11:-10,
 12: -8,
 13: -6,
 14: -4,
 15: -2 };
for (var i in Y.PC_Imm4)
  Y.Imm4_PC[Y.PC_Imm4[i]]=i;
*/

// shorter code:
for (var i=-8; i<8; i++){
  var j=i<<1;
  if (!(j & 8))
    j |= 16;
  Y.PC_Imm4[
    Y.Imm4_PC[j]=i&15
  ]=j;
}
