/* yasep/core-js/yasep_opcodes.js
This file is distributed under terms of the Affero GPL v3 license or later, see http://yasep.org
created 2006-08-08 by whygee@f-cpu.org
2006-08-08
2006-08-16 : added vanity pretty-print
2006-09-02 : moved to vspsim/JScore/
2007-04-06 : some rework for the Queues and the conditions
2007-04-07 : allowing opcode aliases
2007-04-09 : moved decorate_opcode() from JSgui/decoration.js,
   moved the flags away from the forms field
2007-11-05 : rebranding to YASEP
2008-11-30 :
 * removed A0/D0
 * adding 16-bit computation functions
2009-07-19 : added NewAlias()
2009-07-28 : added Y
2011-09-11 : decorate_opcode updated
2011-11-08 : changed the order of A and D register for the new "parking" system
2012-08-13 : added the "check" argument to NewOpcode
2014-02-08 : 2011-11-08 undone

This file defines the general opcode table
(and the reverse tables) that help encoding and
decoding the YASEP instructions.

note : the opcodes themselves are generated in different files.

Dependency : yasep_forms.js for the declaration of Y
*/

Y.keywords={};

/////////////////////////////// The core's registers: ///////////////////////////////
Y.registers_names=[
  "PC", // PC is a very special register... PC=0
  "R1","R2","R3","R4","R5",
  "D1", "A1",
  "D2", "A2", // change LSB to get either A or D
  "D3", "A3",
  "D4", "A4",
  "D5", "A5"
];

// the reverse table :
Y.names_registers=[];
for (var i=0; i<16; i++){
  Y.keywords[Y.registers_names[i]]=true;
  Y.names_registers[Y.registers_names[i]]=i;
}

/////////////////////////////// The instruction object : ///////////////////////////////

Y.prefix=0; // declared and initialized here. Must not be touched later
// or the other instructions won't initialize correctly !!! (except for alignment)
//Y.opcode_step=32;//1<<5;

Y.opcode_table=[]; // contains all the important data
Y.table_opcode=[]; // reverse table for the above one
Y.opcode_aliases=[]; // List of aliases and their "real" names


// build both tables at once
Y.NewOpcode=function(name, action32, action16, forms, flags, description, check) {
  // the variable "prefix" must already exist
  if (Y.table_opcode[name]!=undefined) {
    alert(I8N("woad")+name+I8N("woad2"));
    return;
  }
  Y.keywords[name/*.toUpperCase()*/]=true;
  Y.table_opcode[name]={
    val: Y.prefix, // value of the alias
    forms: forms,  // the forms of the alias may be different from the base's forms.
    flags: flags,
    description: description,
    check:check};
  Y.opcode_table[Y.prefix]={
    name:name,
    action32:action32,
    action16:action16,
    forms:forms,
    flags:flags,
    description:description};
  Y.prefix+=32;//Y.opcode_step; // increment inside the group

  if (window.winman)
    winman.register_key("ISM/"+name,ISM.start);
};

Y.NewAlias=function(name, aliasTo, forms, flags, description) {
  if (window.winman)
    winman.register_key("ISM/"+name,ISM.start);
  Y.NewOpcode(name, null, null, forms, flags, description);
  Y.opcode_aliases[name]=aliasTo;
  Y.prefix-=32;//Y.opcode_step; // leave the counter unchanged so the real instruction can come after.
};

// build both tables at once
Y.NoOpcode=function() {
  Y.prefix+=32;//Y.opcode_step; // increment inside the group
};

// pretty-print :
Y.decorate_opcode=function(opcode){
  var t=opcode, d=Y.table_opcode, h;
  if (typeof opcode == "number"){
    t=Y.opcode_table[opcode].name;
    d=Y.opcode_table;
  }

  h='<a onclick="wa(\'ISM/'+t+'\')" href="#!ISM/'+t+'" class="';
  if (Y.opcode_aliases[t])
     h+="alias_active";
  else
     h+="opcode_active";
  return h+'" title="'+d[opcode].description+'">'+t+"</a>";
}
